import axios from "axios";
import { KeyDataPair, MessageData } from "../types";
import { vereign } from "../generated/qrcode_data_pb";

import {
  arrayBufferToBase64,
  base64ToArrayBuffer,
  ensureUint8Array,
  decompressData,
  arrayBufferToHex,
} from "../utils/common";

import CloudflareService from "./CloudflareService";
import CryptoServiceDefault from "./CryptoService";
import { ICryptoService } from "./CryptoService/ICryptoService";
import { CryptoService } from "../index";

const EmailDataMessageV1 = vereign.protobuf.qrcode_data.EmailData_V1;
const KeyDataMessageV1 = vereign.protobuf.qrcode_data.KeyData_V1;
const WrapperDataMessage = vereign.protobuf.qrcode_data.WrapperData;

const EMAIL_DATA_V1 = "EmailData_V1";

interface WrappedData {
  version: number;
  data: ArrayBuffer;
  className: string;
}

const encodeEmailData = (emailData: MessageData): string => {
  const emailDataMessage = EmailDataMessageV1.fromObject(emailData);
  const emailDataBuffer = EmailDataMessageV1.encode(emailDataMessage).finish();

  const wrappedData = {
    version: 1,
    className: EMAIL_DATA_V1,
    data: emailDataBuffer,
  };

  const wrappedDataMessage = WrapperDataMessage.fromObject(wrappedData);
  const wrappedDataBuffer = WrapperDataMessage.encode(
    wrappedDataMessage
  ).finish();
  return arrayBufferToBase64(wrappedDataBuffer);
};

const decodeEmailData = (
  binary: string | Uint8Array | ArrayBuffer
): MessageData => {
  const wrappedDataMessage = WrapperDataMessage.decode(
    ensureUint8Array(binary)
  );
  const wrappedData = WrapperDataMessage.toObject(
    wrappedDataMessage
  ) as WrappedData;

  if (wrappedData.className === EMAIL_DATA_V1) {
    const emailDataMessage = EmailDataMessageV1.decode(
      ensureUint8Array(wrappedData.data)
    );

    const messageData = EmailDataMessageV1.toObject(emailDataMessage, {
      bytes: String,
      arrays: true, // populates empty arrays (repeated fields) even if defaults=false
      objects: true, // populates empty objects (map fields) even if defaults=false
    }) as MessageData;

    return messageData;
  } else {
    return null;
  }
};

const encodeKeyDataPair = (keyData: KeyDataPair): string => {
  const keyDataMessage = KeyDataMessageV1.fromObject(keyData);
  const buffer = KeyDataMessageV1.encode(keyDataMessage).finish();

  return arrayBufferToBase64(buffer);
};

const decodeKeyDataPair = (
  binary: string | Uint8Array | ArrayBuffer
): KeyDataPair => {
  const keyDataPairMessage = KeyDataMessageV1.decode(ensureUint8Array(binary));
  return KeyDataMessageV1.toObject(keyDataPairMessage, {
    bytes: String,
  }) as KeyDataPair;
};

const breakQrCodeData = (
  data: ArrayBuffer,
  headBytesSize = 32
): { head: ArrayBuffer; tail: ArrayBuffer } => {
  if (data.byteLength <= headBytesSize) {
    return {
      head: data,
      tail: null,
    };
  } else {
    const dataUint8Array = new Uint8Array(data);
    return {
      head: dataUint8Array.slice(0, headBytesSize).buffer,
      tail: dataUint8Array.slice(headBytesSize, dataUint8Array.length).buffer,
    };
  }
};

const assembleQrCodeData = (
  head: ArrayBuffer,
  tail: ArrayBuffer
): ArrayBuffer => {
  return Buffer.concat([Buffer.from(head), Buffer.from(tail)]);
};

const computeQrCodeHash = async (emailData: MessageData): Promise<string> => {
  let attachments = [];
  if (emailData.attachments) {
    attachments = emailData.attachments.map((attachment) => {
      //TODO: check if we can use attachment hash, attachment hashAlg
      if (attachment.url) {
        return {
          name: attachment.name,
          size: attachment.size,
          url: attachment.url,
        };
      }

      return {
        name: attachment.name,
        size: attachment.size,
      };
    });
  }

  const dataForHashing = {
    sender: emailData.sender,
    subject: emailData.subject,
    recipients: emailData.recipients,
    attachments,
  };

  const promises = Object.values(dataForHashing).map((value) => {
    const string = Buffer.from(
      JSON.stringify(value, (key, value) =>
        value instanceof Object && !(value instanceof Array)
          ? Object.keys(value)
              .sort()
              .reduce((sorted, key) => {
                sorted[key] = value[key];
                return sorted;
              }, {})
          : value
      )
    ).toString();

    return CryptoService.SHA256(string);
  });

  const hashArray = await Promise.all(promises);

  const hashesAsAstring: string = hashArray
    .map(arrayBufferToBase64)
    .sort()
    .join("\n");

  const resultBuffer = await CryptoService.SHA256(hashesAsAstring);

  return arrayBufferToBase64(resultBuffer);
};

const calculateQRCodeSignature = async (
  accountPrivateKey: string,
  qrCodeHash: string
): Promise<string> => {
  const signature = await CryptoService.signRSA(
    accountPrivateKey,
    Buffer.from(qrCodeHash)
  );

  return arrayBufferToBase64(signature);
};

const verifyQrCodeSignature = async (
  publicKey: string,
  qrCodeSignature: string,
  recipientQRCodeHash: string
): Promise<boolean> => {
  const qrCodeSignatureAsBuffer = base64ToArrayBuffer(qrCodeSignature);
  const recipientQRCodeHashAsBuffer = Buffer.from(recipientQRCodeHash);
  const result = await CryptoService.verifyRSASignature(
    publicKey,
    recipientQRCodeHashAsBuffer,
    qrCodeSignatureAsBuffer
  );

  return result;
};

interface WithServices {
  getMessageDataFromBase64: (base64: string) => Promise<MessageData>;
}

interface BackblazeDataPartRaw {
  key_signature: string;
  qr_code_data: string;
  session_key: string;
}

/**
 * Higher order function to perform complex operations with the QR code data
 * @param cloudFlareServiceInstance = new CloudflareService("CDN_RUL", "DEFAULT_BUCKET")
 * @param landingPageServiceUrl - URL of the service responsible for RSA decryption
 * @param cryptoService - instance of ICryptoService. Override only needed cryptographic functions.
 * The rest will be filled with default ones from CryptoServiceWeb/CryptoServiceNode
 */
const withServices = (
  cloudFlareServiceInstance: CloudflareService,
  landingPageServiceUrl: string,
  cryptoService?: ICryptoService
): WithServices => {
  if (cryptoService) {
    cryptoService = { ...CryptoServiceDefault, ...cryptoService };
  } else {
    cryptoService = CryptoServiceDefault;
  }

  return {
    getMessageDataFromBase64: async (base64: string) => {
      const qrCodeDataPart = decodeKeyDataPair(base64);

      const base64SHA256 = await cryptoService.SHA256(base64);

      const {
        data: backblazeDataPart,
      } = await cloudFlareServiceInstance.fetchFile<BackblazeDataPartRaw>(
        `qrcode-${arrayBufferToHex(base64SHA256)}`
      );

      const assembledData = assembleQrCodeData(
        base64ToArrayBuffer(qrCodeDataPart.data),
        base64ToArrayBuffer(backblazeDataPart.qr_code_data)
      );

      const {
        key: aesEncryptedSessionKey,
        data: storageIv,
      } = decodeKeyDataPair(backblazeDataPart.session_key);

      const rsaEncryptedSessionKey = await cryptoService.decryptAESGCM(
        base64ToArrayBuffer(aesEncryptedSessionKey),
        base64ToArrayBuffer(qrCodeDataPart.key),
        base64ToArrayBuffer(storageIv)
      );

      const {
        data: { key: encodedSessionKey },
      } = await axios({
        url: `${landingPageServiceUrl}/api/hsm/decrypt`,
        method: "POST",
        headers: {
          "Content-Type": "application/json",
        },
        data: {
          key_signature: backblazeDataPart.key_signature,
          encrypted_key: rsaEncryptedSessionKey,
        },
      });

      const decodedSessionKey = decodeKeyDataPair(encodedSessionKey);

      const decryptedEmailData = await cryptoService.decryptAESGCM(
        assembledData,
        base64ToArrayBuffer(decodedSessionKey.key),
        base64ToArrayBuffer(decodedSessionKey.data)
      );

      const decompressedEmailData = decompressData(decryptedEmailData);
      return decodeEmailData(decompressedEmailData);
    },
  };
};

export default {
  encodeEmailData,
  decodeEmailData,
  encodeKeyDataPair,
  decodeKeyDataPair,
  breakQrCodeData,
  assembleQrCodeData,
  computeQrCodeHash,
  calculateQRCodeSignature,
  verifyQrCodeSignature,
  withServices,
};
